/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.anttasks.sitebuilder;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;
import java.util.Iterator;
import java.util.Vector;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Task;

public class SiteBuilder extends Task {

		private File siteFile=null; // location of file that will be created
		private URL mirrorsURL=null;
		private Vector categories=new Vector();
		private Vector archives=new Vector();
		private Description description=null;
		public void setFile(File siteFile) {
			this.siteFile=siteFile;
		}
		
		public void addDescription(Description d) {
			if(description!=null)  throw new BuildException("only one description allowed");
			this.description=d;
		}
		public void addCategory(Category category) {
			categories.add(category);
		}
		public void addArchive(Archive archive) {
			archives.add(archive);
		}
		public void setMirrorsURL(URL url) {
			this.mirrorsURL=url;
		}

		public void execute() throws BuildException {
			try {
				execute1();
			} catch(Exception e) {
				e.printStackTrace();
			}
		}
		public void execute1() throws BuildException {
		
			if(siteFile==null) throw new BuildException("missing site file");
			
			if(categories.isEmpty()) throw new BuildException("no categories specified");
			
			try {
				FileWriter fw=new FileWriter(siteFile);
				PrintWriter pw=new PrintWriter(fw);
				
				// write document header
				
				pw.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
				pw.println("<!-- AUTO GENERATED SITE FILE using "+getClass().getName()+" -->");
				
				// write start site tag
				
				pw.print("<site");
				if(mirrorsURL!=null) {
					pw.print(" mirrorsURL=\""+mirrorsURL.toExternalForm()+"\" ");
				}
				pw.println(">");
				
				// write description
				if(description!=null) {
					
					pw.print("<description");
					String url=description.getUrl();
					if(url!=null) pw.print(" url=\""+url+"\" ");
					pw.println(">");
					pw.println(description.getText().trim());
					pw.println("</description>");
				}
				// write categories
				Iterator categoryIterator=categories.iterator();
				while(categoryIterator.hasNext()) {
					Category category=(Category) categoryIterator.next();
					writeCategoryEntries(pw,category);
				}
				// write categories
				categoryIterator=categories.iterator();
				while(categoryIterator.hasNext()) {
					Category category=(Category) categoryIterator.next();
					writeCategoryHeaders(pw,category);
				}
				// write archives
				writeArchives(pw);
				
				// write end site tag
				
				pw.println("</site>");
				
				pw.close();
			} catch (IOException e) {
				
				throw new BuildException(e);
			}
			
		}

		
		private void writeArchives(PrintWriter pw) {
			Iterator i=archives.iterator();
			while(i.hasNext()) {
				Archive archive=(Archive) i.next();
				String path=archive.getPath();
				String url=archive.getUrl();
				pw.println("<archive path=\""+path+"\" url=\""+url+"\"/>");
			}
			
		}

		private void writeCategoryEntries(PrintWriter pw, Category category) {
			String label=category.getLabel();
			if(label==null) throw new BuildException("category label missing");
			String name=category.getName();
			if(name==null) throw new BuildException("category name missing");
			
			String urlPrefix=category.getUrlPrefix();
			
			Vector files=category.getFiles();
			Iterator fileIterator=files.iterator();
			while(fileIterator.hasNext()) {
				String fileName=(String) fileIterator.next();
				int underscore=fileName.indexOf("_");
				String version=fileName;
				String id=fileName;
				
				
				if(underscore>1) {
					id=fileName.substring(0,underscore);
					version=fileName.substring(underscore+1);
				}
				
				if(version.endsWith(".jar") || version.endsWith(".zip")) {
					version=version.substring(0,version.length()-4);
				}
				pw.println("\t<feature url=\""+urlPrefix+fileName+"\" id=\""+id+"\" version=\""+version+"\">");
				pw.println("\t\t<category name=\""+name+"\"/>");
				pw.println("\t</feature>");
			}
			
		}

		public URL getMirrorsURL() {
			return mirrorsURL;
		}

		private void writeCategoryHeaders(PrintWriter pw, Category category) {
			String label=category.getLabel();
			if(label==null) throw new BuildException("category label missing");
			String name=category.getName();
			if(name==null) throw new BuildException("category name missing");
			
			pw.print("<category-def name=\""+name+"\" label=\""+label+"\"");
			String description=category.getDescriptionText();
			if(description.trim().equals("")==false) {
				pw.println(">");
				pw.println("<description>");
				pw.println(description);
				pw.println("</description>");
				pw.println("</category-def>");
			}
			else {
			pw.println("/>");
			}
			
			
			
		}
		
}
