/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;

import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.ImagePointer;



/**
 * <p>
 * Represents a Java class.
 * <p>
 * A Java Class can have fields and methods. It is a shallow model of a loaded <a href="http://java.sun.com/docs/books/jvms/second_edition/html/ClassFile.doc.html">class file</a>
 * or special types such as array types or primitive types in the Java Virtual Machine.  
 * 
 */
public interface JavaClass {

    /**
     * <p>
     * Fetch the {@link java.lang.Class} object associated with this class.
     * </p><p>
     * In some implementations this may be null if no object has been
     * created to represent this class, or if the class is synthetic.
     * </p>
     * @return the java.lang.Class object associated with this class 
     * @throws CorruptDataException if the underlying data is in an unexpected state 
     * 
     * @see #getID()
     */
    JavaObject getObject() throws CorruptDataException; 

    /**
     * Fetch the class loader associated with this class. Classes defined in
     * the bootstrap class loader (including classes representing primitive 
     * types or void) will always return a JavaClassLoader representing the
     * bootstrap class loader. This asymmetry with 
     * java.lang.Class#getClassLoader() is intentional. 
     * 
     * @throws CorruptDataException if the class loader for this class cannot
     * be found (a class cannot exist without a loader so this implies corruption)
     * 
     * @return the JavaClassLoader in which this class was defined
     */
    JavaClassLoader getClassLoader() throws CorruptDataException;
    
    /**
     * <p>Get the name of the class in a form that follows the {@link java.lang.Class#getName()} definition.</p>
     * <p>This method will always return a valid class name.</p>
     *   
     * @return the name of the class
     * 
     * @throws CorruptDataException  if the underlying data is in an unexpected state
     */
    String getName() throws CorruptDataException;
    
    /**
     * <p>
     * Get the  super class of this class.
     * </p><p>
     * Will return the superclass of this class or null if no superclass exists.
     *</p><p> 
     * For JavaClass instances representing interfaces, java.lang.Object, primitive types ( int,boolean,char etc) 
     * and void, calling this method will return null.
  	 *</p>
     * @return the immediate superclass of this class, or null if this class has no superclass.
     * 
     * @throws CorruptDataException if the underlying data is in an unexpected state
     */
    JavaClass getSuperclass() throws CorruptDataException;
    
    
    /**
     * <p>Get the set of names of interfaces directly implemented by the class represented by this JavaClass.</p>
     *<p>Some JVM implementations may choose to load interfaces lazily, so only the names are returned.</p>
     *<p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned list is never null but could be empty.</p>
     * 
     * @return a list  of the names of interfaces directly implemented by this class. 
     * 
     * @see java.lang.String
     * @see JavaClassLoader#findClass(String)
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<String> getInterfaces();
    
    /**
     * <p>
     * Return the Java language modifiers for this class. 
     * </p><p>
     * The modifiers are defined by the <a href="http://java.sun.com/docs/books/jls/second_edition/html/classes.doc.html#21613">JVM Specification</a>.
     * </p><p>
     * Note that, for inner classes, the actual modifiers are returned, not 
     * the synthetic modifiers. For instance, a class will never have its
     * 'protected' modifier set, even if the inner class was a protected
     * member, since 'protected' is not a legal modifier for a class file.
     * </p>
     * @return the modifiers for this class
     * @throws CorruptDataException if the underlying data is in an unexpected state
     */
    int getModifiers() throws CorruptDataException;
    
    /**
     * This method returns true if the class represented by this JavaClass is an array class.
     * 
     * @return true if this class is an array class
     * @throws CorruptDataException if the underlying data is in an unexpected state
     */
    boolean isArray() throws CorruptDataException;
    
    /**
     * For array classes, returns a JavaClass representing the component type of this array class.
     *
     * @return a JavaClass representing the component type of this array class
     * @throws CorruptDataException if the underlying data is in an unexpected state
     * 
     * @throws java.lang.IllegalArgumentException if this JavaClass does not represent an array class
     */
    JavaClass getComponentType() throws CorruptDataException;
    
    /**
     * <p>Get the set of fields declared in this class.</p>
     * <p>Fields declared in any superclass of this class are not returned.</p>
     * <p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned list is never null but could be empty.</p>
     * 
     * @return a list of fields declared in this class. 
     * 
     * @see JavaField
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<JavaField> getDeclaredFields();
    
    /**
     * <p>Get the set of methods declared in this class.</p>
     * <p>Methods declared in any superclass of this class are not returned.</p>
     * <p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned list is never null but could be empty.</p>
 
     * @return a list of methods declared in this class.
     * 
     * @see JavaMethod
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<JavaMethod> getDeclaredMethods();
    
    /**
     * <p>
     * Returns the list of constant pool references defined by this class.
     * <p>
     * Java classes may refer to other classes and to String objects via 
     * the class's constant pool. These references are followed by the 
     * garbage collector, forming edges on the graph of reachable objects.
     * This getConstantPoolReferences() may be used to determine which
     * objects are referred to by the receiver's constant pool.
     * <p>
     * Although Java VMs typically permit only Class and String 
     * objects in the constant pool, some esoteric or future virtual 
     * machines may permit other types of objects to occur in the constant 
     * pool. This API imposes no restrictions on the types of JavaObjects 
     * which might be included in the list.
     * <p>
     * No assumption should be made about the order in which constant 
     * pool references are returned.
     * <p>
     * Classes may also refer to objects through static variables.
     * These may be found with the getDeclaredFields() API. Objects
     * referenced by static variables are not returned by 
     * getConstantPoolReferences() unless the object is also referenced
     * by the constant pool.
     * 
     * <p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned list is never null but could be empty.</p>
     * 
     * @return a list of JavaObjects which are referred to by the constant pool of this class.
     * 
     * @see JavaObject
     * @see javax.tools.diagnostics.image.CorruptData
     */
    List<JavaObject> getConstantPoolReferences();
    
    /**
     * <p>
     * The ID of a class is a pointer to a section of memory which identifies
     * the class. The contents of this memory are implementation defined.
     * <p>
     * In some implementations getID() and getObject().getID() may return the
     * same value. This implies that the class object is also the primary
     * internal representation of the class. API users should not rely on this
     * behaviour.
     * <p>
     * In some implementations, getID() may return null for some classes.
     * 
     * @return a pointer to the class
     */
	ImagePointer getID();
	
	/**
	 * <p>
	 * Get the set of references from this class. 
	 * <p>
	 * A reference is a object that represents the uni-directional relationship between objects and classes.
	 * Objects and classes cannot be reclaimed by the Java Virtual Machine garbage collector if references exist
	 * that can ultimately be traced back to root references.  see {@link JavaReference} for more detailed information.
	 * <p>
	 * Since this API can present entities that exist at any point in their lifecycle, it is possible to encounter an
	 * JavaClass that is eligible for collection and thus no {@link JavaReference} can be found that refers to it.
	 *   
	 * <p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned list is never null but could be empty.</p>
     * 
     * @return a list of JavaReferences
	 *
	 * @see javax.tools.diagnostics.runtime.java.JavaReference
	 * @see javax.tools.diagnostics.image.CorruptData
	 */
	List<JavaReference> getReferences();
    
	/**
	 * @param obj
	 * @return True if the given object refers to the same Java Class in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
