/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;


import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;



/**
 * <p>
 * Represents an internal ClassLoader structure within a Java Virtual Machine instance.
 * For most ClassLoaders there is a corresponding {@link java.lang.ClassLoader}
 * instance within with JavaRuntime. For primordial class loaders such as
 * the bootstrap class loader, there may or may not be a corresponding
 * {@link java.lang.ClassLoader} instance.
 * </p>
 * <p>
 * Since Java does not define any strict inheritance structure between
 * class loaders, there are no APIs for inspecting 'child' or 'parent'
 * class loaders. This information may be inferred by inspecting the
 * corresponding {@link java.lang.ClassLoader} instance:<p>
 * <i>pseudo javacode example</i>
 * <pre>
 * 		JavaClassLoader  loader;
 * 		JavaObject instance=loader.getObject();
 * 		String classLoaderName=instance.getJavaClass().getName();
 * 
 * </pre>
 * 
 * @see java.lang.ClassLoader
 */
public interface JavaClassLoader {

    /**
     * <p>
     * Get the set of classes which are defined in this JavaClassLoader.
     * Calling the {@link JavaClass#getClassLoader()} method on objects  returned in this list will
     * return this JavaClassLoader</p>
     *    
     * <p>The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned value is never null but can be an empty list.</p>
     *   
     * @return an list of classes which are defined in this JavaClassLoader 
     * @see JavaClass
     */
    List<JavaClass> getDefinedClasses();
    
    /**
     * <p>
     * When a ClassLoader successfully delegates a findClass() request to
     * another ClassLoader, the result of the delegation must be cached within
     * the internal structure so that the Java Virtual Machine does not make repeated requests
     * for the same class.
     * </p>
     * The contents of the returned list is a superset of that returned by {@link #getDefinedClasses()}
     * <p>
     * The returned list follows the standard semantics for javax.tools.diagnostics collections.</p>
     * <p>The returned value is never null but can be an empty list.</p>
     * 
     * @return a list of classes which are defined  in this JavaClassLoader 
     * <i>or</i> which were found by delegation to other JavaClassLoaders
     * 
     * @see JavaClass
     */
    List<JavaClass> getCachedClasses();
    
    /**
     * <p>
     * Find a class by name within this class loader. The class may have been
     * defined in this class loader, or this class loader may have delegated
     * the load to another class loader and cached the result.
     *<p>
     * The form of the name presented to this method should be as follows
     * <pre>
     * [ packagenamepart / ... ] (classname ) [ $innerclassname ...]
     * </pre>
     * <p>
     * <i>Examples</i>
     * </p>
     * <ul>
     * <li>To find the JavaClass that represents "java.lang.String" use findClass("java/lang/String")</li>
     * <li>To find the JavaClass that represents "Foo.InnerClass.InnerInnerClass" in the default package use findClass("Foo$InnerClass$InnerInnerClass")</li>
     * <li>To find the JavaClass that represents "java.util.Map.Entry use findClass("java/util/Map$Entry")
     * </ul> 
     * 
     * @param name of the class to find. Packages should be separated by
     * '/' instead of '.'
     * @return the JavaClass instance, or null if it is not found
     * @throws CorruptDataException if the underlying data is in an unexpected state
     */
    JavaClass findClass(String name) throws CorruptDataException;
    
    /**
     * Get the {@link java.lang.ClassLoader} instance (represented by a {@link JavaObject} 
     * associated with this class loader. If there is no associated class loader, for example the
     * system class loader , then  null will be returned.
     * 
     * Further examination of the returned object is implementation specific.
     *  
     * @return a JavaObject representing the {@link java.lang.ClassLoader} instance
     * 
     * @throws CorruptDataException if the underlying data is in an unexpected state
     * 
     * @see JavaObject
     * @see ClassLoader
     */
    JavaObject getObject() throws CorruptDataException;
    
	/**
	 * @param obj
	 * @return True if the given object refers to the same Java Class Loader in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
