/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;

/**
 * <p>
 * Represents a Java reference.   
 * <p>
 * A Java reference is a traceable relationship between two objects or between a root
 * and a Java object.
 * <p>
 * References are used by Garbage Collection systems to identify objects that can be 
 * reclaimed.
 *  
 */
public interface JavaReference
{
    
	/** Unknown heap root type */
    int HEAP_ROOT_UNKNOWN      = 0;
    /** JNI global reference heap root */
    int HEAP_ROOT_JNI_GLOBAL   = 1;
    /** System class heap root */
    int HEAP_ROOT_SYSTEM_CLASS = 2;
    /** Monitor heap root */
    int HEAP_ROOT_MONITOR      = 3;
    /** Stack local heap root */
    int HEAP_ROOT_STACK_LOCAL  = 4;
    /** JNI local reference heap root */
    int HEAP_ROOT_JNI_LOCAL    = 5;
    /** Thread heap root */
    int HEAP_ROOT_THREAD       = 6;
    /** Other heap root type */
    int HEAP_ROOT_OTHER        = 7;
    /** Finalizable object heap root */
    int HEAP_ROOT_FINALIZABLE_OBJ = 8;
    /** Unfinalized object heap root */
    int HEAP_ROOT_UNFINALIZED_OBJ = 9;
    /** Classloader heap root */
    int HEAP_ROOT_CLASSLOADER = 10;
    /** Stringtable heap root */
    int HEAP_ROOT_STRINGTABLE = 11;
    

     /** Unknown reference type */
    int REFERENCE_UNKNOWN           = 0;
    /** Reference from an object to its class */ 
    int REFERENCE_CLASS             = 1;  // 
    /** Reference from an object to the value of one of its instance fields */
    int REFERENCE_FIELD             = 2;  
    /** Reference from an array to one of its elements */
    int REFERENCE_ARRAY_ELEMENT     = 3;
    /** Reference from a class to its class loader */
    int REFERENCE_CLASS_LOADER      = 4;
    /** Reference from a class to its signers array */
    int REFERENCE_SIGNERS           = 5;
    /** Reference from a class to its protection domain */
    int REFERENCE_PROTECTION_DOMAIN = 6;
    /** Reference from a class to one of its interfaces */
    int REFERENCE_INTERFACE         = 7;
    /** Reference from a class to the value of one of its static fields */
    int REFERENCE_STATIC_FIELD      = 8; 
    /** Reference from a class to a resolved entry in the constant pool */
    int REFERENCE_CONSTANT_POOL     = 9;
    /** Reference from a class to its superclass */
    int REFERENCE_SUPERCLASS     	= 10;
    /** Reference from a classloader object to its loaded classes */
    int REFERENCE_LOADED_CLASS     	= 11;
    /** Reference from a class to its java.lang.Class instance */
    int REFERENCE_CLASS_OBJECT      = 12;
	/** Reference from a JavaObject representing a Class to the associated JavaClass **/
    int REFERENCE_ASSOCIATED_CLASS 	= 13;


    /** Reachability of target object via this reference is unknown */
    int REACHABILITY_UNKNOWN =  0;
    /** Reachability of target object via this reference is Strong */
    int REACHABILITY_STRONG  =  1;
    /** Reachability of target object via this reference is Soft */
    int REACHABILITY_SOFT    =  2;
    /** Reachability of target object via this reference is Weak */
    int REACHABILITY_WEAK    =  3;
    /** Reachability of target object via this reference is Phantom */
    int REACHABILITY_PHANTOM =  4;

    /**
     * <p>
     * Get the root type, as defined in the JVMTI specification.
     *
     * @return an integer representing the root type, see HEAP_ROOT_ statics above.
     */
    int getRootType() throws CorruptDataException;

    /**
     * <p>
     * Get the reference type, as defined in the JVMTI specification.
     * @return an integer representing the reference type, see REFERENCE_ statics above.
     */
    int getReferenceType() throws CorruptDataException;

    /**
     * <p>
     * Get the reachability of the target object via this specific reference.
     * @return an integer representing the reachability, see REACHABILITY_ statics above.
     */
    int getReachability() throws CorruptDataException;

    /**
     * <p>
     * Get a string describing the reference type.
     * <p>
     * Users should not depend on the contents or identity of this string.
     * e.g. "JNI Weak global reference", "Instance field 'MyClass.value'", "Constant pool string constant"
     * @return a String describing the reference type
     */
    String getDescription();

    /**
     * <p>
     * Check to see if this reference points to an object in the heap
     * @return true if the target of this root is an object
     * @throws DataUnavailable if the requested information is not available
     * @throws CorruptDataException is the underlying data is in an unexpected state
     */
    boolean isObjectReference() throws DataUnavailable, CorruptDataException;

    /**
     * <p>
     * Check to see if  this reference points to a class.
     * 
     * @return true if the target of this root is a class
     * @throws DataUnavailable if the requested information is not available
     * @throws CorruptDataException is the underlying data is in an unexpected state
     */
    boolean isClassReference() throws DataUnavailable, CorruptDataException;

    /**
     * <p>
     * Get the object referred to by this reference.
     * @return a JavaObject or a JavaClass
     * @throws DataUnavailable if the requested information is not available
     * @throws CorruptDataException is the underlying data is in an unexpected state
     */
    Object getTarget() throws DataUnavailable, CorruptDataException;

    /**
     * <p>
     * Get the source of this reference if available.
     * @return a JavaClass, JavaObject, JavaStackFrame, JavaThread or null if unknown
     * @throws DataUnavailable if the requested information is not available
     * @throws CorruptDataException is the underlying data is in an unexpected state
     */
    Object getSource() throws DataUnavailable, CorruptDataException;
}
