/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;


import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;



/**
 * <p>Represents a Java stack frame.</p>
 * 
 *
 */
public interface JavaStackFrame {
    /**
     * <p>Get a pointer to the base of this stack frame within memory.</p>
     * <p>The layout of a <code>JavaStackFrame</code> is implementation specific.
     * </p>
     * 
     * @return the base pointer of the stack frame
     * @throws CorruptDataException 
     */
    ImagePointer getBasePointer() throws CorruptDataException;
    
	/**
     * <p>
     * Returns the {@link JavaLocation} that represents the location of this 
     * <code>JavaStackFrame</code> within the Java program.
     * </p>
     * 
     * @return a location object describing where the frame is executing.
	 * @throws CorruptDataException 
     * 
     * @see JavaLocation
     */
    JavaLocation getLocation() throws CorruptDataException;
    
    /**
     * <p>A list of references to objects and classes from this stack frame.
     * </p>
	 * <p>
	 * Returns the references to object and classes this Java Virtual Machine's implementation
	 * considers as being kept alive by this Java stack frame. As well as references from local variables and
	 * operations stack entries, this may also include a reference to the Java frame's class or to
	 * an object this stack frame is keeping alive through holding its monitor.
	 * </p>
	 * 
     * @return a list of JavaReferences
     */
	List<JavaReference> getHeapRoots();                      

	/**
	 * 
	 * <p>Gets the value of a variable from a stack frame.</p>
	 * 
	 * <p>Returns a {@link JavaObject} for an object reference, <code>null</code> for a null object reference.
	 * Primitives are returned as boxed primitives.
	 * CorruptDataException is thrown if object reference is incorrect, or if the float or double are set to invalid values.
	 * </p>
	 * 
	 * @param slot - the numerical local variable slot number to retrieve.
	 * 
	 * @throws DataUnavailable if this method is not supported or if stack not in correct state to return variables.
	 * @throws IndexOutOfBoundsException if an invalid slot number is passed.
	 * 
	 * @experimental this addition to the API is experimental and subject to change
	 */

	Object getVariable(int slot) throws CorruptDataException, DataUnavailable, IndexOutOfBoundsException;

	/**
	 * <p>Gets all variables from the stack frame.</p>
	 * <p>
	 * Returns a {@link java.util.List} containing the available {@link JavaVariable JavaVariables}.
	 * While the method {@link JavaMethod#getVariables()} will return all local variables for a method,
	 * this method will return only those variables that are visible at the point of execution for this
	 * stack frame.
	 * </p>
	 *@experimental this addition to the API is experimental and subject to change
	 * 
	 */

	List<JavaVariable> getVariables(); 

	/**
	 * @param obj
	 * @return True if the given object refers to the same JavaStackFrame in the image
	 */
	public boolean equals(Object obj);
	public int hashCode();
}
