/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package javax.tools.diagnostics.runtime.java;

import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;



/**
 * This class models the JavaVMInitArgs C structure passed to <a href="http://java.sun.com/javase/6/docs/technotes/guides/jni/spec/invocation.html#wp16334">JNI_CreateJavaVM</a>
 * to create this Java Virtual Machine
 * 
 * Typically the options passed to the JVM are similar but necessarily identical to these used to invoke the Java Virtual Machine from a command line.
 * 
 */
public interface JavaVMInitArgs {

	/**
	 * The JNI specified version constant for the Java 1.1 version of JNI 
	 */
	public static final int JNI_VERSION_1_1 = 0x00010001;

	/**
	 * The JNI specified version constant for the Java 1.2 version of JNI
	 */
	public static final int JNI_VERSION_1_2 = 0x00010002;

	/**
	 * The JNI specified version constant for the Java 1.4 version of JNI
	 */
	public static final int JNI_VERSION_1_4 = 0x00010004;
	
	/**
	 * The JNI specified version constant for the Java 1.6 version of JNI
	 */
	public static final int JNI_VERSION_1_6 = 0x00010006;

	/**
	 * Fetch the JNI version from the JavaVMInitArgs structure used to create this Java Virtual Machine.
	 * See the <a href="http://java.sun.com/javase/6/docs/technotes/guides/jni/spec/functions.html#wp15951">JNI specification</a> for the meaning for this field.
	 *  
	 * @return the JNI version
	 * @throws DataUnavailable
	 * @throws CorruptDataException
	 */
	int getVersion() throws DataUnavailable, CorruptDataException;
	
	/**
	 * Fetch the ignoreUnrecognized field from the JavaVMInitArgs structure used to create this Java Virtual Machine.
	 * See the <a href="http://java.sun.com/javase/6/docs/technotes/guides/jni/spec/functions.html#wp15951">JNI specification</a> for the meaning for this field.
	 * 
	 * @return true if ignoreUnrecognized was set to a non-zero value when the Java Virtual Machine was invoked
	 * @throws DataUnavailable
	 * @throws CorruptDataException
	 */
	boolean getIgnoreUnrecognized() throws DataUnavailable, CorruptDataException;
	
	/**
	 * Fetch the options used to start this Java Virtual Machine, in the order they were originally specified.
	 * 
	 * The returned list follows the standard semantics for javax.tools.diagnostics collections
	 * 
	 * The order of the options returned in the list is the same as that passed to the to <a href="http://java.sun.com/javase/6/docs/technotes/guides/jni/spec/invocation.html#wp16334">JNI_CreateJavaVM</a> function.
	 * A list is always returned but could be empty
	 * 
	 * @return an List of JavaVMOptions
	 * @throws DataUnavailable
	 * 
	 * @see JavaVMOption
	 */
	List<JavaVMOption> getOptions() throws DataUnavailable;

}
