/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaClass;
import javax.tools.diagnostics.runtime.java.JavaClassLoader;
import javax.tools.diagnostics.runtime.java.JavaField;
import javax.tools.diagnostics.runtime.java.JavaObject;

import org.apache.kato.common.BasicType;
import org.apache.kato.hprof.IJavaClass;
import org.apache.kato.hprof.datalayer.ConstantPoolEntry;
import org.apache.kato.hprof.datalayer.FieldEntry;
import org.apache.kato.hprof.datalayer.InstanceFieldEntry;
import org.apache.kato.hprof.datalayer.StaticFieldEntry;
import org.apache.kato.hprof.image.CorruptDataImpl;
import org.apache.kato.hprof.image.ImagePointerImpl;

/**
 * HProf implementation of JavaClass. 
 *  
 *
 */
public class JavaClassImpl implements JavaClass {
	IJavaHeapInternal heap;
	IJavaClass javaClass;
	
	/**
	 * Superclass to static and instance fields.
	 *
	 */
	abstract class JavaFieldImpl implements JavaField {
		FieldEntry field;

		/**
		 * Returns the signature of this field.
		 * Unfortunately, only the fact that a field is an object reference
		 * is recorded. 
		 * If we find all objects an object reference field refers to,
		 * if we take the lowest common subclass then that is the closest to
		 * the field's actual type. It would be prohibitively expensive to do
		 * that though.
		 */
		@Override
		public String getSignature() throws CorruptDataException {
			switch (field.getType()) {
			case BasicType.OBJECT:
				return "Ljava/lang/Object;";
			case BasicType.BOOLEAN:
				return "Z";
			case BasicType.CHAR:
				return "C";
			case BasicType.FLOAT:
				return "F";
			case BasicType.DOUBLE:
				return "D";
			case BasicType.BYTE:
				return "B";
			case BasicType.SHORT:
				return "S";
			case BasicType.INT:
				return "I";
			case BasicType.LONG:
				return "J";
			default:
				throw new CorruptDataException(new CorruptDataImpl("Unexpected basic type of " + field.getType()+
						"encountered."));
			}
		}

		@Override
		public String getName() throws CorruptDataException {
			return heap.getUTF8StringByID(field.getFieldNameID());
		}
		
		@Override
		public JavaClass getDeclaringClass() throws CorruptDataException,
				DataUnavailable {
					return JavaClassImpl.this;
				}
		
		String getStringFromJavaObject(JavaObject obj) throws CorruptDataException, MemoryAccessException {
			JavaClass clazz = obj.getJavaClass();
			JavaObject value=null;
			int offset=0;
			int count=0;
			boolean valueSet = false, offsetSet = false, countSet=false;
			
			Iterator iter = clazz.getDeclaredFields().iterator();
			
			while (iter.hasNext()) {
				JavaField field = (JavaField) iter.next();
				String name = field.getName();
				if("value".equals(name)) {
					value = (JavaObject) field.get(obj);
					valueSet = true;
				} else if ("offset".equals(name)) {
					offset = field.getInt(obj);
					offsetSet = true;
				} else if ("count".equals(name)) {
					count = field.getInt(obj);
					countSet = true;
				}
				
			}

			if (valueSet == false || offsetSet == false || countSet == false ) {
				throw new CorruptDataException(new CorruptDataImpl("Unable to find all String classes fields values"));
			}
			char characters[] = new char[count];
			
			value.arraycopy(offset, characters, 0, count);
			
			return new String(characters);
		}
		
		/**
		 * Checks the compatibility of this field with a return type.
		 * Only call from a specific type. 
		 * Will always throw an exception if called against an object
		 * reference field.
		 * 
		 * @param type the type this field is to be returned as
		 * @throws IllegalArgumentException if type is incompatible
		 */
		void checkType(int type) {
			switch(field.getType()) {
			case BasicType.OBJECT:
				throw new IllegalArgumentException("objcet field can only be returned as a object.");
			case BasicType.BOOLEAN:
				if(type != BasicType.BOOLEAN) {
					throw new IllegalArgumentException("boolean field can only be returned as a boolean.");
				}
				break;
			case BasicType.BYTE:
				if (type == BasicType.BOOLEAN) {
					throw new IllegalArgumentException("byte field can not be returned as a boolean.");
				}
				if(type == BasicType.CHAR) {
					throw new IllegalArgumentException("byte field can not be returned as a char.");
				}
				break;
			case BasicType.SHORT:
				switch (type) {				
				case BasicType.BOOLEAN:				
				case BasicType.BYTE:
				case BasicType.CHAR:	
					throw new IllegalArgumentException("short field can not be returned as a smaller type.");
				}
				break;
			case BasicType.CHAR:
				switch (type) {
				case BasicType.BOOLEAN:				
				case BasicType.BYTE:
				case BasicType.SHORT:
					throw new IllegalArgumentException("char field can not be returned as a smaller type.");
				}
				break;
			case BasicType.INT:
				switch (type) {
				case BasicType.BOOLEAN:				
				case BasicType.BYTE:
				case BasicType.SHORT:
				case BasicType.CHAR:
					throw new IllegalArgumentException("int field can not be returned as a smaller type.");
				}
				break;
				
			case BasicType.LONG:
				switch (type) {
				case BasicType.BOOLEAN:				
				case BasicType.BYTE:
				case BasicType.SHORT:
				case BasicType.CHAR:
				case BasicType.INT:
					throw new IllegalArgumentException("long field can not be returned as a smaller type.");
				}
				break;
			case BasicType.FLOAT:
				switch (type) {
				case BasicType.BOOLEAN:				
				case BasicType.BYTE:
				case BasicType.SHORT:
				case BasicType.CHAR:
				case BasicType.INT:
				case BasicType.LONG:
					throw new IllegalArgumentException("float field can not be returned as a smaller type.");
				}
				break;
				
			case BasicType.DOUBLE:
				switch (type) {
				case BasicType.BOOLEAN:				
				case BasicType.BYTE:
				case BasicType.SHORT:
				case BasicType.CHAR:
				case BasicType.INT:
				case BasicType.LONG:
				case BasicType.FLOAT:
					throw new IllegalArgumentException("double field can not be returned as a smaller type.");
				}
				break;
			}
				
		}
	}
	
	class InstanceFieldImpl extends JavaFieldImpl {
		
		
		public InstanceFieldImpl(InstanceFieldEntry field) {
			this.field = field;
		}
		
		/**
		 * Returns true if this field can retrieve a value from the passed class.
		 * 
		 * @param clazz JavaClassImpl of object to retrieve field from
		 * @return true if the class is compatible
		 * 
		 * @throws CorruptDataException
		 */
		private boolean isCompatible(JavaClassImpl clazz) throws CorruptDataException {
			// checks that this field can be extracted from the passed clazz
			
			while(clazz != null) {
				if (clazz.equals(JavaClassImpl.this)) {
					return true;
				}
				
				clazz = clazz.getSuperclass();
			}
			
			return false;
		}
		
		/** 
		 * Returns the offset of this field into a given object.
		 * 
		 * @param clazz JavaClassImpl
		 * @return int offset
		 * @throws CorruptDataException
		 */
		private int getOffset(JavaClassImpl clazz) throws CorruptDataException {
			// The potential subclasses of this field have to be skipped before
			// using the stored field offset.		
			int offset = ((InstanceFieldEntry)field).getOffset();			
			while ((clazz != null) && !clazz.equals(JavaClassImpl.this)) {
				offset += clazz.getOffsetSize(); // Or some variant.
				clazz = clazz.getSuperclass();
			}
			return offset;
		}
		
		@Override
		public Object get(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			
			if (obj == null) {
				throw new NullPointerException("Instance JavaField passed null object.");
			}
			if (! (obj instanceof JavaObjectInstanceImpl)) {
				throw new IllegalArgumentException("Passed object must be an ordinary object");			
			}
			
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			
			checkObjectCompatible(obj);
 			
			// The potential subclasses of this field have to be skipped before
			// using the stored field offset. 
			int offset = getOffset(clazz);
			
			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);

			switch(field.getType()) {
			case BasicType.BOOLEAN:
				return new Boolean(instance.getBooleanField(offset));
			case BasicType.BYTE: 
				return new Byte(instance.getByteField(offset));
			case BasicType.SHORT: 
				return new Short(instance.getShortField(offset));
			case BasicType.CHAR: 
				return new Character(instance.getCharField(offset));
			case BasicType.INT: 
				return new Integer(instance.getIntegerField(offset));
			case BasicType.LONG: 
				return new Long(instance.getLongField(offset));
			case BasicType.FLOAT: 
				return new Float(instance.getFloatField(offset));
			case BasicType.DOUBLE: 
				return new Double(instance.getDoubleField(offset));
			case BasicType.OBJECT:
				return JavaClassImpl.this.heap.getObjectByID(instance.getIDField(offset));
			default:
				throw new CorruptDataException(new CorruptDataImpl("Invalid field type "
						+field.getType()));
			}
		}

		/**
		 * Check if the passed obj is compatible with this class.
		 * if not, throw IllegalArgumentException.
		 * @param obj JavaObject to test against
		 * @throws CorruptDataException 
		 * @throws IllegalArgumentException
		 */
		private void checkObjectCompatible(JavaObject obj) throws CorruptDataException {
			if (!(obj instanceof JavaObjectInstanceImpl)) {
				throw new IllegalArgumentException("Passed object not an instance of JavaObjectInstanceImpl. "+obj.getClass().getName());
			}
			if (!isCompatible((JavaClassImpl) obj.getJavaClass())) {
				throw new IllegalArgumentException("Field " + 
						JavaClassImpl.this.getName() + "." + 
						getName() +" not in class hierachy of object 0x"+
						Long.toHexString(obj.getID().getAddress()));
			}
		}
		
		@Override
		public boolean getBoolean(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.BOOLEAN);
			checkObjectCompatible(obj);

			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			
			checkObjectCompatible(obj);
			
			int offset = getOffset(clazz);
			
			return instance.getBooleanField(offset);
		}


		@Override
		public byte getByte(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.BYTE);
			checkObjectCompatible(obj);

			
			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			int offset = getOffset(clazz);
			
			return instance.getByteField(offset);
		}

		@Override
		public char getChar(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.CHAR);
			checkObjectCompatible(obj);

			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			int offset = getOffset(clazz);
			
			return instance.getCharField(offset);
		}

		@Override
		public double getDouble(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.DOUBLE);
			checkObjectCompatible(obj);

			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			int offset = getOffset(clazz);
			
			return instance.getDoubleField(offset);
		}

		@Override
		public float getFloat(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.FLOAT);
			checkObjectCompatible(obj);

			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			int offset = getOffset(clazz);
			
			return instance.getFloatField(offset);

		}

		@Override
		public int getInt(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.INT);
			checkObjectCompatible(obj);

			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			int offset = getOffset(clazz);
			
			return instance.getIntegerField(offset);

		}

		@Override
		public long getLong(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.LONG);
			checkObjectCompatible(obj);

			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			int offset = getOffset(clazz);
			
			return instance.getLongField(offset);

		}

		@Override
		public short getShort(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.SHORT);
			checkObjectCompatible(obj);

			JavaObjectInstanceImpl instance = ((JavaObjectInstanceImpl) obj);
			JavaClassImpl clazz = (JavaClassImpl) obj.getJavaClass();
			int offset = getOffset(clazz);
			
			return instance.getShortField(offset);
		}

		@Override
		public String getString(JavaObject obj) throws CorruptDataException,
				MemoryAccessException {
			checkObjectCompatible(obj);

			Object fieldObj = get(obj);
			if (fieldObj == null) {
				return null;
			}
			if (!(fieldObj instanceof JavaObject)) {
				throw new IllegalArgumentException("Field "+getName()+" not an object");
			}
			JavaObject jObject = (JavaObject) fieldObj;
			
			if (! "java/lang/String".equals( jObject.getJavaClass().getName() )) {
				throw new IllegalArgumentException("Field "+getName()+" refers to 0x"+
						Long.toHexString(jObject.getID().getAddress()) +
						", which not a string - it's a " + jObject.getJavaClass());
			}
			
			return getStringFromJavaObject(jObject);
		}

		@Override
		public int getModifiers() throws CorruptDataException {
			return 0;
		}
		
	}
	
	/**
	 * Represents a static field.
	 * Passed JavaObject is always ignored. The actual value is stored
	 * within the StaticFieldEntry.
	 * 
	 */
	class StaticFieldImpl extends JavaFieldImpl {

		StaticFieldImpl(StaticFieldEntry field) {
			this.field = field;
		}
		
		/**
		 * Retrieves the field and returns an Object - primitive types are
		 * boxed. Other types are implementations of JavaObject, or null.
		 */
		@Override
		public Object get(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			StaticFieldEntry field = (StaticFieldEntry) this.field;
			
			switch(field.getType()){
			case BasicType.OBJECT:
				return heap.getObjectByID(field.getIDField());
			case BasicType.BOOLEAN:
				return new Boolean(field.getBooleanField());
			case BasicType.CHAR:
				return new Character(field.getCharField());
			case BasicType.FLOAT:
				return new Float(field.getFloatField());
			case BasicType.DOUBLE:
				return new Double(field.getDoubleField());
			case BasicType.BYTE:
				return new Byte(field.getByteField());
			case BasicType.SHORT:
				return new Short(field.getShortField());
			case BasicType.INT:
				return new Integer(field.getIntegerField());
			case BasicType.LONG:
				return new Long(field.getLongField());
			default:
				throw new CorruptDataException(new CorruptDataImpl("Unexpected basic type of " + field.getType()+
						"encountered."));		
			}
		}

		@Override
		public boolean getBoolean(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.BOOLEAN);

			return ((StaticFieldEntry)field).getBooleanField();
		}

		@Override
		public byte getByte(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.BYTE);
			return ((StaticFieldEntry)field).getByteField();
		}

		@Override
		public char getChar(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.CHAR);
			return ((StaticFieldEntry)field).getCharField();
		}

		@Override
		public double getDouble(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.DOUBLE);
			return ((StaticFieldEntry)field).getDoubleField();
		}

		@Override
		public float getFloat(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.FLOAT);
			return ((StaticFieldEntry)field).getFloatField();
		}

		@Override
		public int getInt(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.INT);

			return ((StaticFieldEntry)field).getIntegerField();
		}

		@Override
		public long getLong(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.LONG);

			return ((StaticFieldEntry)field).getLongField();
		}

		@Override
		public short getShort(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			checkType(BasicType.SHORT);

			return ((StaticFieldEntry)field).getShortField();
		}

		@Override
		public String getString(JavaObject arg0) throws CorruptDataException,
				MemoryAccessException {
			
			return getStringFromJavaObject(heap.getObjectByID(((StaticFieldEntry)field).getIDField()));
		}

		/**
		 * This field is static - nothing else is known.
		 */
		@Override
		public int getModifiers() throws CorruptDataException {
			return java.lang.reflect.Modifier.STATIC;
		}
	}
	
	public JavaClassImpl(IJavaHeapInternal heap, IJavaClass javaClass) {
		this.heap = heap;
		this.javaClass = javaClass;
	}
	
	@Override
	public JavaClassLoader getClassLoader() throws CorruptDataException {
		return heap.getJavaClassLoaderByID(javaClass.getClassLoaderObjectID());
	}

	/**
	 * Get the component type of this class if it is an array class.
	 * 
	 * Primitive arrays have primitives as elements, so their component
	 * type are classes called "int", etc. These classes are faked using 
	 * JavaPrimitiveClassImpl as they don't exist in hprof dumps. 
	 *  
	 * An int[][] array is an array of arrays, hence not primitive.
	 * 
	 * The component types are retrieved by name from the JavaClassLoader this
	 * array class belongs to. 
	 */
	@Override
	public JavaClass getComponentType() throws CorruptDataException {
		if (!isArray()) {
			throw new IllegalArgumentException("Class is not an array type");
		}
		
		String name = javaClass.getClassName();
		
		if ("boolean[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.BOOLEAN);
		}if ("byte[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.BYTE);
		}if ("short[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.SHORT);
		}if ("char[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.CHAR);
		}if ("int[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.INT);
		}if ("long[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.LONG);
		}if ("float[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.FLOAT);
		}if ("double[]".equals(name)) {
			return heap.getPrimitiveClass(BasicType.DOUBLE);
		}
		
		String componentName = this.getName().substring(1);
		
		/* If the array was "[Ljava/lang/String;"
		 * Strip it to "java/lang/String"
		 */
		if(componentName.charAt(0) == 'L') {
			componentName = componentName.substring(1,componentName.length()-1);
		}
		
		return getClassLoader().findClass(componentName);
	}

	@Override
	public List getConstantPoolReferences() {
		List cpeList = new LinkedList();
		
		
		ConstantPoolEntry[] cpes = this.javaClass.getConstantPoolEntries();

		for ( ConstantPoolEntry cpe : cpes) {
			if (cpe.type == BasicType.OBJECT) {
			//	cpeList.add(heap.getObjectByID(cpe.data));
			}
		}
		return null;
	}

	/**
	 * The static and instance fields belonging to this class.
	 * 
	 */
	private ArrayList<JavaField> fields;
	@Override
	public List getDeclaredFields() {
		if (fields == null) {
			StaticFieldEntry staticFields[] = javaClass.getStaticFields();
			InstanceFieldEntry instanceFields[] = javaClass.getInstanceFields();
			
			fields = new ArrayList<JavaField>(staticFields.length + instanceFields.length);
			
			for (StaticFieldEntry f : staticFields) {
				fields.add(new StaticFieldImpl(f));
			}
			
			for (InstanceFieldEntry f : instanceFields) {
				fields.add(new InstanceFieldImpl(f));
			}
		}
		
		return fields;
	}

	/** 
	 * Method information isn't stored within classes.
	 * 
	 */
	@Override
	public List getDeclaredMethods() {
		return new LinkedList();
	}

	
	/**
	 * Wraps up object ID of class as it's ID.
	 * That should be sufficiently unique.
	 * 
	 * @return ImagePointer
	 */
	@Override	
	public ImagePointer getID() {
		return heap.getImagePointer(javaClass.getClassObjectID());
	}

	/**
	 * Implemented interfaces aren't stored in Hprof.
	 * The constant pool could be used if the right entries could be targeted, 
	 * but there is nothing to suggest that that is possible.
	 */
	@Override
	public List getInterfaces() {
		return new LinkedList();
	}

	/**
	 * There is no modifier information to return
	 * @return 0
	 */
	@Override
	public int getModifiers() throws CorruptDataException {
		return 0;
	}

	/**
	 * Returns the classes's name. This is stored in hprof with "." and "$" separators,
	 * the "."s are replaced by "/"s to conform with the API.
	 * 
	 * @return String of the class's name.
	 */
	@Override
	public String getName() throws CorruptDataException {
		String name = javaClass.getClassName();
		
		/* Process the array's such that they are of the format:
		 * [Ljava/lang/Object;
		 * 
		 * or 
		 * [[B
		 *  
		 * Primitive arrays are renamed correctly.
		 */
		if(name.indexOf('[') != -1) {
			int depth=0;
			String prefix = "";
			for(int n=0; n < name.length(); n++) {
				if (name.charAt(n) == '[') {
					prefix+="[";
				}
			}
			
			// Chop off arrays.
			name = name.substring(0, name.indexOf('['));
			
			if (name.startsWith("boolean")) {
				name = "Z";
			} else if (name.startsWith("byte")) {
				name = "B";
			} else if (name.startsWith("char")) {
				name = "C";
			} else if (name.startsWith("short")) {
				name = "S";
			} else if (name.startsWith("int")) {
				name = "I";
			} else if (name.startsWith("long")) {
				name = "J";
			} else if (name.startsWith("float")) {
				name = "F";
			} else if (name.startsWith("double")) {
				name = "D";
			} else {
				name = "L" + name.replace('.', '/') +";";
			}
			return prefix+name;
		}
		return name.replace('.', '/');
	}

	/**
	 * Returns null. There are no java/lang/Class objects referenced
	 * by the class records. 
	 */
	@Override
	public JavaObject getObject() throws CorruptDataException {
		return null;
	}

	/**
	 * Return the superclass of this class or null.
	 * 
	 * @return JavaClassImpl or null
	 */
	@Override
	public JavaClassImpl getSuperclass() throws CorruptDataException {
		long superClassObjId = javaClass.getSuperClassObjectID();
		
		// This is an assumption, but it will be necessary.
		if (superClassObjId == 0L) {
			return null;
		}
		
		return heap.getJavaClassByID(superClassObjId);
	}

	@Override
	public List getReferences() {
		// TODO Auto-generated method stub
		// All object references, constantpool entries, superclass, etc.
		return null;
	}

	/**
	 * Returns true if this is an array.
	 * 
	 * @return true if this class is an array type.
	 */
	@Override
	public boolean isArray() throws CorruptDataException {
		return javaClass.getClassName().indexOf('[') != -1;
	}

	/**
	 * @return true if passed object is this JavaClass
	 */
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		} 
		
		if (!(obj instanceof JavaClassImpl)) {
			return false;
		}
		
		if (obj == this) {
			return true;
		}
		
		JavaClassImpl clazz = (JavaClassImpl) obj;
		
		return clazz.getObjectID() == getObjectID();
	}
	
	/**
	 * Gets the object ID that uniquely identifies this class on the heap.
	 * 
	 * @return long object ID
	 */
	long getObjectID() {
		return javaClass.getClassObjectID();
	}
	
	public String toString() {
		String name = "<UNSET CLASS NAME>";
		try {
			name = this.getName();
		} catch (CorruptDataException e) {
			// oh well.
		}
		return name+" @ "+ Long.toHexString(javaClass.getClassObjectID());
	}
	
	public int hashCode() {
		return (int) (javaClass.getClassObjectID() % Integer.MAX_VALUE);
	}
	
	/**
	 * Return the size of object instances.
	 * 
	 * @return int of the object size.
	 */
	int getInstanceSize() {
		return javaClass.getInstanceSize();
	}
	
	/**
	 * Returns the number of bytes the part of this class takes up in an instance object.
	 * This is the number of bytes taken up by the fields in this class and none of its
	 * superclasses.
	 * This is different from getInstanceSize(), which is the size of an instance of this
	 * class, including all superclasses.
	 * 
	 * @return int fields size
	 */
	int getOffsetSize() {
		return javaClass.getOffsetSize();
	}
}
