/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.hprof.java;

import java.util.LinkedList;
import java.util.List;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.image.DataUnavailable;
import javax.tools.diagnostics.image.ImagePointer;
import javax.tools.diagnostics.image.MemoryAccessException;
import javax.tools.diagnostics.runtime.java.JavaHeap;
import javax.tools.diagnostics.runtime.java.JavaObject;

import org.apache.kato.hprof.datalayer.IGCInstanceHeapDumpRecord;

final class JavaObjectInstanceImpl implements JavaObject {
	/**
	 * 
	 */
	private final IGCInstanceHeapDumpRecord record;
	private IJavaHeapInternal heap;
	
	JavaObjectInstanceImpl(IJavaHeapInternal javaHeapImpl, IGCInstanceHeapDumpRecord record) {
		this.record = record;
		this.heap = javaHeapImpl;
	}

	@Override
	public void arraycopy(int srcStart, Object dst,
			int dstStart, int length)
			throws CorruptDataException,
			MemoryAccessException {
		throw new IllegalArgumentException("instance is not an array");		
	}

	@Override
	public int getArraySize()
			throws CorruptDataException {		
		throw new IllegalArgumentException("instance is not an array");
	}

	@Override
	public long getHashcode() throws DataUnavailable,
			CorruptDataException {		
		return record.getID();
	}

	@Override
	public JavaHeap getHeap()
			throws CorruptDataException,
			DataUnavailable {
		return heap;
	}

	@Override
	public ImagePointer getID() {		
		return heap.getImagePointer(record.getID());
	}

	@Override
	public JavaClassImpl getJavaClass()
			throws CorruptDataException {
		return heap.getJavaClassByID(record.getClassObjectID());
	}

	@Override
	public long getPersistentHashcode()
			throws DataUnavailable,
			CorruptDataException {
		throw new DataUnavailable("Persistent hashcode unknown.");
	}

	@Override
	public List getReferences() {
		return new LinkedList();
	}

	@Override
	public List getSections() {		
		return new LinkedList();
	}

	@Override
	public long getSize() throws CorruptDataException {
		return getJavaClass().getInstanceSize(); 
	}

	@Override
	public boolean isArray()
			throws CorruptDataException {
		return false;
	}

	/**
	 * Retrieve a byte field from this object.
	 * 
	 * @param offset into instance values.
	 * @return byte value
	 */
	boolean getBooleanField(int offset) {
		return record.getBooleanField(offset);
	}
	
	/**
	 * Retrieve a byte field from this object.
	 * 
	 * @param offset into instance values.
	 * @return byte value
	 */
	byte getByteField(int offset) {
		return record.getByteField(offset);
	}
	
	/**
	 * Retrieve a char field from this object.
	 * 
	 * @param offset into instance values.
	 * @return char value
	 */
	char getCharField(int offset) {
		return record.getCharField(offset);
	}
	
	/**
	 * Retrieve a short field from this object.
	 * 
	 * @param offset into instance values.
	 * @return short value
	 */
	short getShortField(int offset) {
		return record.getShortField(offset);
	}
	
	/**
	 * Retrieve a int field from this object.
	 * 
	 * @param offset into instance values.
	 * @return int value
	 */
	int getIntegerField(int offset) {
		return record.getIntegerField(offset);
	}
	
	/**
	 * Retrieve a long field from this object.
	 * 
	 * @param offset into instance values.
	 * @return long value
	 */
	long getLongField(int offset) {
		return record.getLongField(offset);
	}
	
	/**
	 * Retrieve a float field from this object.
	 * 
	 * @param offset into instance values.
	 * @return float value
	 */
	float getFloatField(int offset) {
		return record.getFloatField(offset);
	}
	
	/**
	 * Retrieve a double field from this object.
	 * 
	 * @param offset into instance values.
	 * @return double value
	 */
	double getDoubleField(int offset) {
		return record.getDoubleField(offset);
	}
	
	/**
	 * Retrieve an object ID field from this object.
	 * 
	 * @param offset into instance values.
	 * @return long value - an object ID.
	 */
	long getIDField(int offset) {
		return record.getIDField(offset);
	}
	
	/**
	 * Compares identity. This object may be created multiple times.
	 * Uses object ID for comparing identity. 
	 * We don't check runtime identity.
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		
		if (obj instanceof JavaObjectInstanceImpl) {
			JavaObjectInstanceImpl other = (JavaObjectInstanceImpl) obj;
			
			if (other.getObjectID() == record.getID()) {
				return true;
			}
		}
		
		return false;
	}
	
	public long getObjectID() {
		return record.getID();
	}
	
}