/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tools.jdi;
import java.io.PrintStream;
import java.net.BindException;
import java.net.InetSocketAddress;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.Vector;

/* JDISocketServer
 * 
 * JDISocketServer is the server that accepts new clients and spawns a new
 * JDISocket for them.  We hold a weak reference to the JDWPServer, and pass
 * a strong reference to the JDISocket (which then turns around and creates
 * a new WeakReference).
 * 
 * We also hold a vector of JDISockets, so when someone calls unlisten(), we
 * tell all the JDISockets to disconnect(), and then close our own connection,
 * which causes us to drop out and shut down.
 * 
 * We use new io (nio) for the socket stuff, so that we can get access to select()
 * in JDISocket.
 * 
 */

public class JDISocketServer extends Thread{
	private Vector<JDISocket> vctr = new Vector<JDISocket>();
	private boolean shuttingDown = false;
	private ServerSocketChannel svr;
	private JDWPServer jdwpServer;
	private int nPort;
	private boolean terminateOnLastClient = false; 
	
	public JDISocketServer(JDWPServer jdwpserver, int port) throws Exception{
		jdwpServer = jdwpserver;
		nPort = port;
		try{
			svr = ServerSocketChannel.open();
			svr.socket().bind(new InetSocketAddress (nPort));
			logr.log(JDILogger.LEVEL_VERBOSE, "Listening on " + nPort); //$NON-NLS-1$
		}
		catch (Exception exxy){
			throw exxy;
		}
	}

	public void setTerminateLastClient(boolean setting){
		terminateOnLastClient = setting;
	}

	public boolean getTerminateLastClient(){
		return terminateOnLastClient;
	}

	private JDILogger logr = JDILogger.logger;

	public void unlisten(){
		shuttingDown = true;

		for(int i = 0; i < vctr.size(); i++){
			synchronized(vctr){
				JDISocket vsock = vctr.elementAt(i);
				try{
					vsock.disconnect();
				}catch(Exception exxy){ }
				vctr.remove(vsock);
			}
		}

		try{
			svr.close();
		}catch(Exception exxy){ }	
		svr = null;
		jdwpServer = null;

	}
	
	@Override
	public void run(){
		try{
			while(true){
				SocketChannel schan = svr.accept();
				JDISocket jdiSocket = new JDISocket(jdwpServer, schan, this);
				synchronized(vctr){
					vctr.add(jdiSocket);
				}
				//Look for stale references
				processConnections();


			}

		}catch(BindException exxy){
			return;
		}catch(java.nio.channels.AsynchronousCloseException exxy){ 
			if (!shuttingDown){
				exxy.printStackTrace();
				return;
			}
		}catch(Exception exxy){
			logr.logError(JDILogger.LEVEL_NORMAL, "Unable to start server " + exxy.getMessage()); //$NON-NLS-1$
			exxy.printStackTrace(new PrintStream(logr.getErrorStream()));
			return;
		}
		for(int i = 0; i < vctr.size(); i++){
			JDISocket vsock = vctr.elementAt(i);
			try{
				vsock.disconnect();
			}catch(Exception exxy){ }
			vctr.remove(vsock);
		}
	}

	/**
	 * notifyDisconnect is called by the JDISocket when it shuts down.
	 * You can think of it like a 'hook', so we can count how many connections
	 * we have open right now, and shut down the server if we need to.
	 */
	public void notifyDisconnect(){
		processConnections();
		synchronized(vctr){
			if (!shuttingDown){
				if (vctr.size() < 1 && terminateOnLastClient){
					jdwpServer.stop();
				}
			}
		}
	}
	
	/**
	 * addConnection forces the addition of sock to the connection list. 
	 * @param sock JDISocket object to add
	 */
	public void addConnection(JDISocket sock){
		vctr.add(sock);
	}

	/**
	 * processConnections rescans the JDISocket vector, and if a socket
	 * isn't connected, shut it down.
	 */
	private void processConnections(){
		synchronized(vctr){
			for(int i = 0; i < vctr.size(); i++){
				if (!vctr.elementAt(i).isConnected()){
					vctr.remove(i);
				}
			}
		}
	}

}
