/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/

/****

CJVMTI v0.1
-> indicates what this can possibly call 
addThreads -> addObject & addClass
addObject -> addClass & addObject & queues objects
addClass -> addClass & queues classes
objectQueues -> getObjectReferences
getObjectReferences -> addField
addField -> addObject
classQueue -> getStaticReferences
getStaticReferences -> addObject

addObject and addClass both can function without following any references and allocate the required space in the dump file
This means calling them can happen at any point with no adverse effects (running out of stack.)
They are tagged so that any references to them are resolved to their position in the dump file.
Its only once the dump is completed that the references are followed by clearing the class and object queue. These both generate
many more references that need to be added to the queues. This two step process means all reference loops are easily avoided as references
can always be written as the reference is only a pointer inside the file.

****/

#include "cjvmti.h"
#include "addThreads.h"
#include "addField.h"
#include "addClass.h"
#include "addObject.h"

int jvmtiVersion = 11;
int dumping = 0; // Used to ensure two dumps aren't attempted at the same time
static int logFile = 0; 
int initAgent();

// Provides printing to log/stdout
void printfR(const char * _Format, ...) {
	va_list args;
	va_start(args, _Format);
	//vprintf(_Format, args);

	va_end(args);
	va_start(args, _Format);
	if (logFile) {
		//vfprintf(fp, _Format, args);
	}
	va_end(args);
}

// Provides printing to log/stdout
void printfd(const char * _Format, ...) {
	int i;
	va_list args;
	va_start(args, _Format);
	assert(printDepth > -1);

	for (i = 0; i < printDepth; i++) {
		//printf(" ");
		if (logFile) {
			//fprintf(fp, " ");
		}
	}
	//vprintf(_Format, args);

	va_end(args);
	va_start(args, _Format);
	if (logFile) {
		//vfprintf(fp, _Format, args);
	}
	va_end(args);
}

int writeIDNoSize(char id, FILE * fp) {
	fwrite(&id, sizeof(char), 1, fp);
	return 1;
}

// Writes a null terminated string, needs replacing to specifying the number of characters
int writeString(char * tString, FILE * fp) {
	fputs(tString, fp);
	fputc(0, fp);
	return 1;
}

int writeJint(jint i, FILE * fp) {
	fwrite(&i, sizeof(jint), 1, fp);
	return 1;
}

// Used for writing longs and file position, needs splitting up
int writeReference(FPOS_T pt, FILE * fp) {
	fwrite(&pt, sizeof(FPOS_T), 1, fp);
	return 1;
}

int writeLong(jlong j, FILE *fp) {
	fwrite(&j, sizeof(jlong), 1, fp);
	return 1;
}

int writeBool(jboolean b, FILE * fp) {
	fwrite(&b, sizeof(char), 1, fp);
	return 1;
}

void JNICALL
Exception(jvmtiEnv *jvmti_env,
		  JNIEnv* jni_env,
		  jthread thread,
		  jmethodID method,
		  jlocation location,
		  jobject exception,
		  jmethodID catch_method,
		  jlocation catch_location){
			  jvmtiPhase phas;
			  char * sig;
			  char * gSig;
			  jclass clazz;

			  if (catch_method == NULL){
				  if (!dumping){
					  dumping = 1;
					  printf("Exception %d \n \n \n", catch_method);
					  dataDumpRequest(jvmti_env);
				  }
			  }/*else{ // For apache tomcat demo.
				  if (exception != NULL ){
					  clazz = (*jni_env)->GetObjectClass(jni_env, exception);
					  (*env)->GetClassSignature(env, clazz, &sig, &gSig);
					  if (!strcmp(sig, "Ljava/lang/ArrayIndexOutOfBoundsException;")){
						  if (!dumping){
							  dumping = 1;
							  printf("Exception %d \n \n \n", catch_method);
							  dataDumpRequest(jvmti_env);
						  }
					  }
				  }
			  }*/
			  return;
}

// Clears through reference queue
int clearObjectQueue() {
	jclass clazz;
	jobject obj;
	obj = 0;
	clazz = 0;
	while (Q_dequeue(&obj, qObjects) != 0  || Q_dequeue(&clazz, qStaticFields) != 0) {
		if (obj != 0){
			getObjectReferences(obj);
		}
		if(clazz != 0){
			getClassReferences(clazz);
		}
	}
	return 1;
}

// Ensures references are within the files range, useful for testing
int checkReference(FPOS_T loc, FILE * fp) {
	FPOS_T fileSize;
	FPOS_T currentLoc = FTELL(fp);
	FSEEK(fp, 0, SEEK_END);
	fileSize = FTELL(fp);
	if (loc > fileSize + 1) {
		printf("File size %lld \n", fileSize);
		printf("Bad reference! %lld \n", loc);
		abort();
		return 0;
	}
	FSETPOS(fp, &currentLoc);
	return 1;
}

static void dump() {
	printf("Finished writing dump file \n");
}

static void JNICALL dataDumpRequest(jvmtiEnv *jvmti) {
	int i;
	jint rc;
	jthread currentThread = 0;
	jlong tag;
	jobject deTagObj;
	struct JObject * deAlObj;
	struct JClass * deAlClass;
	jthread * suspendedThreads;
	jint numSuspendedThreads;
	env = jvmti;
	dumping = 1;
	if (jniptr == NULL){
		rc = (*vmptr)->GetEnv(vmptr, (void **) &jniptr, JNI_VERSION_1_4);
		if (rc != JVMTI_ERROR_NONE) {
			printf("Unable to load JNI environment");
			return;
		}
	}

	(*jniptr)->PushLocalFrame(jniptr, 5000); // Automatically deallocates local references when popped

	nullObject.positionInFile = CJVMTI_NULL_OBJECT;
    systemLoader = NULL; /* null system class loader */

	initAgent(); // Prepare dump file


	printf("Get current thread \n");
	(*jvmti)->GetCurrentThread(jvmti, &currentThread);


	(*env)->SetTag(env, currentThread, 1);
	
	if (currentThread == 0){
		printf("Null thread returned \n");
	}else{
	}
	(*env)->SetTag(env, currentThread, 1); // So we can skip pausing the agent thread
	rc = (*env)->GetAllThreads(env, &numSuspendedThreads, &suspendedThreads);
	printf("Suspend the VM threads \n");
	printf("Current thread is %llx \n", currentThread);
	for (i = 0; i < numSuspendedThreads; i++) {
		(*env)->GetTag(env, suspendedThreads[i], &tag);
		if (tag == 0) {
			(*env)->SuspendThread(env, suspendedThreads[i]);
		} else {
			(*env)->SetTag(env, suspendedThreads[i], 0);
		}
	}
	(*env)->SetTag(env,currentThread, 0);

	printf("Creating dump \n");
	qObjects = Q_initQueue(100, sizeof(jobject));
	qStaticFields = Q_initQueue(100, sizeof(jclass));
	toDetag = Q_initQueue(200, sizeof(jobject));
    qClassLoaders = Q_initQueue(10, sizeof(struct JClassLoader*));
	followAllThreads();

    /* Fixup references to classes within classloaders and then deallocate them.
     */
    {
        struct JClassLoader *loader;
        while(Q_dequeue(&loader, qClassLoaders) != 0) {
            writeClassLoader(loader);
        }
    }
	// Dump finished

	if (fp!=NULL)fclose(fp);
	fp = NULL;
	fclose(variableFile);
    variableFile = NULL;

	printf("Detag/deallocate\n");
	while (Q_dequeue(&deTagObj, toDetag)!= 0){
		(*env)->GetTag(env, deTagObj, &tag);
		if (tag == 0)continue; // We sometimes free an object from earlier in the queue if it went from an object to a class object
		deAlObj = (struct JObject *)tag;
		if (deAlObj->type == CLASS){
			deAlClass = (struct JClass *) tag;

			for (i = 0; i < deAlClass->numInstanceFields; i++){
				(*env)->Deallocate(env, (unsigned char*)deAlClass->instanceFieldGenSig[i]);
				(*env)->Deallocate(env, (unsigned char*)deAlClass->instanceFieldSignatures[i]);
			}
			(*env)->Deallocate(env, (unsigned char*)deAlClass->name);
			free(deAlClass->instanceFieldGenSig);
			free(deAlClass->instanceFields);
			free(deAlClass->instanceFieldSignatures);
			free(deAlClass->instanceModifiers);


		}
		free((void*)tag);
		(*env)->SetTag(env, deTagObj, 0);
	}

	printf("Resume threads \n");
	for (i = 0; i < numSuspendedThreads; i++) {
		(*env)->ResumeThread(env, suspendedThreads[i]);
	}

	Q_free(toDetag);
	Q_free(qObjects);
	Q_free(qStaticFields);
	Q_free(qClassLoaders);

    /* system loader will be freed by writeClassLoader. */
    systemLoader = NULL;

	(*env)->Deallocate(env, (unsigned char *)suspendedThreads);
	printf("Finished creating dump file \n");
	(*jniptr)->PopLocalFrame(jniptr, NULL); // Clears local references
	dumping = 0;
}

int setEnvironment(JavaVM * vm){
	jint rc;
	jvmtiError err;
	char * errorName;
	jvmtiCapabilities caps;
	jvmtiEventCallbacks callbacks;
	errorName = NULL;
	rc = (*vm)->GetEnv(vm, (void **) &env, JVMTI_VERSION_1_1);
	if (rc != JNI_OK) {
		printf("--- Cannot connect to JNI ---\n");
		if(rc == JNI_EVERSION){
			printf(" Incompatible version \n");
		}
		return rc;
	}
	memset(&callbacks, 0, sizeof(callbacks));
	callbacks.DataDumpRequest = &dataDumpRequest;
	callbacks.Exception = &Exception;

	err = (*env)->SetEventCallbacks(env, &callbacks, sizeof(callbacks));
	if (err != JVMTI_ERROR_NONE)
		printf("Could not set callbacks");


	err = (*env)->SetEventNotificationMode(env, JVMTI_ENABLE,
			JVMTI_EVENT_DATA_DUMP_REQUEST, (jthread) NULL);

	memset(&caps, 0, sizeof(jvmtiCapabilities));

	caps.can_get_owned_monitor_info = 1;
	caps.can_get_current_contended_monitor = 1;
	caps.can_get_monitor_info = 1;
	caps.can_access_local_variables = 1;
	caps.can_tag_objects = 1;
	caps.can_generate_exception_events = 1;
	caps.can_get_source_file_name = 1;
	caps.can_get_line_numbers = 1;
	caps.can_signal_thread = 1;
	caps.can_suspend = 1;

	err = (*env)->AddCapabilities(env, &caps);
	if (err != JVMTI_ERROR_NONE)
		printf("can't turn on required capabilites\n");

	err = (*env)->SetEventNotificationMode(env, JVMTI_ENABLE,
		JVMTI_EVENT_EXCEPTION, (jthread) NULL);

	if (err != JVMTI_ERROR_NONE)
		printf("Could not set event notification %d \n", err);


	err = (*env)->CreateRawMonitor(env, "cbmonitor", &lock);
	if (err != JVMTI_ERROR_NONE)
		printf("cannot create monitor");
	fp = NULL;
	variableFile = NULL;
	return 1;
}


int initAgent(){
	static int dumpCount;
	time_t creationTime;
	INT64 longCreationTime;
	struct tm  * dumpTime;
	char fileName[50];
	int nameType = 0; // Switches between timestamping the file and not
	int endianCheck = 1;
	objectCount = 0;
	staticsCount = 0;
	printDepth = 0;
	logFile = 0;

	#ifdef __DEBUG_CJVMTI__
		_RPT0(_CRT_WARN,"Trace\n");
		_CrtDumpMemoryLeaks();
	#endif

	if (logFile) {
		printf("Open log file\n");
		fp = fopen("CJVMTI_debug.txt", "wb+");
		fprintf(fp, "CJVMTI LOG FILE\n");
		printf("Opened\n");
	}

	printf("Creating dump file ");
	creationTime = time(NULL);
	dumpTime = localtime(&creationTime);
	if (nameType){
		sprintf(fileName, "cjvmti%04d.%02d.%02d.%02d%02d.%02d.cdmp", 1900+dumpTime->tm_year, 1+dumpTime->tm_mon, dumpTime->tm_mday, dumpTime->tm_hour, dumpTime->tm_min, dumpTime->tm_sec);
	}else{
		sprintf(fileName, "cjvmtiV%d.cdmp", dumpCount);
	}
	printf(fileName);
	printf("\n");
	variableFile = fopen(fileName, "wb+");

	// print header info
	writeReference(0, variableFile);
	fwrite(&endianCheck, sizeof(int), 1, variableFile);
	fprintf(variableFile, "CJVMTI V0.01");

	longCreationTime = (INT64) creationTime; // Ensure that time is written as a long long.
	fwrite(&longCreationTime, sizeof(INT64), 1, variableFile);
	dumpCount++;
	return 1;
}

JNIEXPORT jint JNICALL
Agent_OnLoad(JavaVM *vm, char *options, void *reserved) {
	jint rc;
	jvmtiError err;
	jvmtiCapabilities caps;
	jvmtiEventCallbacks callbacks;
	char * frameDepthChar;
	char * referenceDepthChar;


	printf("CJVMTI v0.1 \n");

	// print header info

	vmptr = vm;
	rc = (*vm)->GetEnv(vm, (void **) &env, JVMTI_VERSION);
	if (rc != JVMTI_ERROR_NONE) {
		printf("--- Cannot connect to JNI ---\n");
		return rc;
	}
	memset(&callbacks, 0, sizeof(callbacks));
	callbacks.DataDumpRequest = &dataDumpRequest;

	err = (*env)->SetEventCallbacks(env, &callbacks, sizeof(callbacks));
	if (err != JVMTI_ERROR_NONE)
		printf("Could not set callbacks");


	err = (*env)->SetEventNotificationMode(env, JVMTI_ENABLE,
			JVMTI_EVENT_DATA_DUMP_REQUEST, (jthread) NULL);
	if (err != JVMTI_ERROR_NONE)
		printf("Could not set event notification");

	memset(&caps, 0, sizeof(jvmtiCapabilities));

	caps.can_get_owned_monitor_info = 1;
	caps.can_get_current_contended_monitor = 1;
	caps.can_get_monitor_info = 1;
	caps.can_access_local_variables = 1;
	caps.can_tag_objects = 1;
	caps.can_generate_exception_events = 1;
	caps.can_get_source_file_name = 1;
	caps.can_get_line_numbers = 1;
	caps.can_signal_thread = 1;
	caps.can_suspend = 1;

	err = (*env)->AddCapabilities(env, &caps);
	if (err != JVMTI_ERROR_NONE)
		printf("can't turn on required capabilites\n");

	err = (*env)->CreateRawMonitor(env, "cbmonitor", &lock);
	if (err != JVMTI_ERROR_NONE)
		printf("cannot create monitor");
	fp = NULL;
	variableFile = NULL;
#ifdef __DEBUG_CJVMTI__
   hLogFile = CreateFile("c:\\memLeakCJVMTI.txt", GENERIC_WRITE,
      FILE_SHARE_WRITE, NULL, CREATE_ALWAYS,
      FILE_ATTRIBUTE_NORMAL, NULL);
   _CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_FILE);
   _CrtSetReportFile(_CRT_WARN, hLogFile);
	_CrtSetDbgFlag ( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
	_CrtDumpMemoryLeaks();
#endif

	if (options != NULL){
		frameDepthChar = strtok(options, ",");
		referenceDepthChar = strtok(NULL, ",");
		if (frameDepthChar == NULL || referenceDepthChar == NULL)
		{
			maxReferenceDepth = 0;
			maxFrameDepth = 0;
			printf("Unlimited frame depth and reference depth set\n");
			return JNI_OK;
		}

		maxFrameDepth = atoi(frameDepthChar);
		maxReferenceDepth = atoi(referenceDepthChar);
		if (maxFrameDepth != 0 || maxReferenceDepth != 0)
		{
			printf("Frame Depth %s ReferenceDepth %s \n", frameDepthChar, referenceDepthChar);
		}else{
			printf("Unlimited frame depth and reference depth set\n");
		}
	}else{
		maxReferenceDepth = 0;
		maxFrameDepth = 0;
		printf("Unlimited frame depth and reference depth set\n");
	}
	return JNI_OK;
}

JNIEXPORT void JNICALL
Agent_OnUnload(JavaVM *vm) {
#ifdef __DEBUG_CJVMTI__
	_RPT0(_CRT_WARN,"file message\n");
	CloseHandle(hLogFile);
#endif
	if (fp != NULL) fclose(fp);
	if (variableFile != NULL) fclose(variableFile);
	//(*env)->DisposeEnvironment(env);

	printf("Agent released \n");
}
