/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.tck.results.reporter;


import java.io.BufferedReader;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Iterator;
import java.util.List;
import java.util.TreeMap;

public class TCKResultMatrixHTMLWriter extends TCKResultMatrixWriter {

	public static final String RESOURCES_STYLESHEET_CSS = "/resources/stylesheet.css";
	public static final String RESOURCES_PURPLE_PNG = "/resources/purple.png";
	public static final String RESOURCES_RED_PNG = "/resources/red.png";
	public static final String RESOURCES_BLUE_PNG = "/resources/blue.png";
	private static final String BR = "<br>";
	private boolean verbose=false;
	
	
	public TCKResultMatrixHTMLWriter(File dir) {
		super(dir);
		
		
	}
	public void init() throws IOException{
		dir.mkdir(); 
		writeStyleSheet(dir);
		writeBinFile(dir,RESOURCES_BLUE_PNG);
		writeBinFile(dir,RESOURCES_RED_PNG);
		writeBinFile(dir,RESOURCES_PURPLE_PNG);
	}


	private void writeStyleSheet(File targetDir) throws IOException {
		
		writeTextFile(targetDir,RESOURCES_STYLESHEET_CSS);
		
		
		
	}

	private void writeTextFile(File targetDir, String name) throws IOException {
		File target=new File(targetDir,name);
		target.getParentFile().mkdirs();
		
		InputStream in=getClass().getResourceAsStream(name);
		if(in==null) {
			log("can't find file "+name);
			return;
		}
		log("writing "+name+" to "+target.getAbsolutePath());	
		FileWriter fw=new FileWriter(target);
		PrintWriter pw=new PrintWriter(fw);
		InputStreamReader ist=new InputStreamReader(in);
		BufferedReader br=new BufferedReader(ist);
		
		while(true) {
			String line=br.readLine();
			if(line==null) break;
			pw.println(line);
		}
		
		br.close();
		pw.close();
	}
	private void writeBinFile(File targetDir, String name) throws IOException {
		
		File target=new File(targetDir,name);
		target.getParentFile().mkdirs();
		
		InputStream in=getClass().getResourceAsStream(name);
		if(in==null) {
			log("can't find file "+name);
			return;
		}
		
		log("writing "+name+" to "+target.getAbsolutePath());
		
		OutputStream  out=new FileOutputStream(target);
		
		byte[] buffer=new byte[10000];
		while(true) {
			int read=in.read(buffer);
			if(read<1) break;
			out.write(buffer,0,read);
		}
		
		out.close();
		in.close();
		
		
	}
	private void log(String msg) {
		System.out.println(msg);
		
	}
	protected void writeDetailFiles(TCKResultMatrix matrix) throws IOException {
		
		File detail=new File(dir,"detail");
		boolean b=detail.mkdirs();
		log("Creating results dir "+detail.getAbsolutePath()+" returned "+b);
		
		SortKeyProvider provider=matrix.getKeyProvider();
		
		Iterator i=matrix.getTCKResultSets();
		
		while(i.hasNext()) {
			TCKResultSet set=(TCKResultSet) i.next();
			String key=provider.getkey(set.getConfig());
			
			File file=new File(detail,key+".html");
			File parent=file.getParentFile();
			b=parent.mkdirs();
			log("creating results child dir "+parent+" (returned "+b+")");
			
			writeStyleSheet(parent);
			
			FileWriter fw=new FileWriter(file);
			PrintWriter pw=new PrintWriter(fw);
			HTMLWriter out=new HTMLWriter(pw);
			
			out.writeDocHeader("detailed output");
			
			out.startDetailTable("standard output");
			out.startRow();
			out.startTableCell();
			out.writeCode(set.getOutLog());
			out.endTableCell();
			
			out.writeEndTable();
			
			out.startDetailTable("standard error");
			out.startRow();
			out.startTableCell();
			out.writeCode(set.getErrLog());
			out.endTableCell();
			out.writeEndTable();
			
			Iterator ri=set.iterator();
			while(ri.hasNext()) {
			TestcaseResult result=(TestcaseResult) ri.next();
				
				String anchor=out.createLink("#"+result.getLabel(),result.getLabel() );
				out.writeHead2(anchor);
				
				if(result instanceof TestcaseResultError) {
					TestcaseResultError err=(TestcaseResultError) result;
					String[] stack=err.getErrorStack();
					out.startDetailTable("Error stack");
					for(int j=0;j<stack.length;j++) {
						String s=stack[j];
						out.startRow();
						out.writeTableCell(s);
						out.endRow();
					}
					out.writeEndTable();
				}
				else if(result instanceof TestcaseResultFailure) {
					TestcaseResultFailure err=(TestcaseResultFailure) result;
					String[] stack=err.getFailureStack();
					out.startDetailTable("Failure stack");
					for(int j=0;j<stack.length;j++) {
						String s=stack[j];
						out.startRow();
						out.writeTableCell(s);
						out.endRow();
					}
					out.writeEndTable();
				}
				
				
				
			}
			out.writeDocEnd();
			
			pw.close();
		}
		
		
	}

	private void writeIndex(TCKResultMatrix matrix) throws IOException {
		
		SortKeyProvider provider=matrix.getKeyProvider();
		
		FileWriter fw=new FileWriter(new File(dir,"index.html"));
		PrintWriter out=new PrintWriter(fw);
		HTMLWriter writer=new HTMLWriter(out);
		writer.writeDocHeader("TCK Results Comparision");
		
		writer.startDetailTable();
		
		// class="Failure"
		
		TreeMap map=new TreeMap();
		
		// how many header rows.
		int depth=matrix.getkeyDepth();
		
		// for each one work out total number of leaf children
		// this gives span of table cell for heading
		
		
		for(int d=0;d<depth;d++) {
			// get values for root headings
			writer.startRow();
			writer.writeTableHeaderCell("");
			
		Key[] keyValues=matrix.getKeys(d);
			
			for(int k=0;k<keyValues.length;k++) {
			Key value=keyValues[k];
				int width=matrix.getKeyColumnWidth(value);
				writer.writeSpanningtableHeader(width,value.getName());
				
				
			}
			
			writer.endRow();	
				
		}
		
		// summary
		writer.startRow();
		writer.writeTableCell("");
		
	
		Key[] keyValues=matrix.getKeys(depth-1);
		if(verbose) log(""+keyValues.length+" columns");
		for(int k=0;k<keyValues.length;k++) {
			Key value=keyValues[k];
			String background="";
			
			String fails="";
			
			int testcases=matrix.getTestcaseCount(value);
			int failures=matrix.getTestcaseFailureCount(value);
			int errors=matrix.getTestcaseErrorCount(value);
			int bad=failures+errors;
			if(bad>0) {
				background="class=\"Failure\" ";
				
				fails=" ("+bad+")";
			}
			String data=""+testcases+fails;
			if(bad>0) {
				
				data="<a href=\""+value.getPath()+"\">"+data+"</a>";
			}
			
			out.print("<td "+background+">"+data+"</td>");
		}
		
		// get leaf node width for each value
		
		writer.endRow();	
		
		String[] rowNames=matrix.getRowNames();
		if(verbose) log(""+rowNames.length+" rows");
		for(int j=0;j<rowNames.length;j++) {
			String rowName=rowNames[j];
		
			writer.startRow();
			writer.writeTableCell(rowName);
			
			for(int k=0;k<keyValues.length;k++) {
				Key value=keyValues[k];
				String fails="";
				TestcaseResult r=matrix.getTestcaseResult(value, rowName);
				String result="";
				if(r!=null) {
				 result=r.getType();
				}
				String imageName=null;
				
				if(result.equals("failed")) {
					fails="class=\"Failure\" ";
					imageName="resources/purple.png";
					
				} else if(result.equals("error")) {
					fails="class=\"Error\" ";
					imageName="resources/red.png";
				}
				else if(result.equals("passed")) {
					imageName="resources/blue.png";
				}
				String link=result;
				TCKResultSet parentSet=matrix.getResultSet(value,rowName);
				if(parentSet!=null) {
					TestExecutionConfig config=parentSet.getConfig();
					
					
					String key=provider.getkey(config);
					String path="detail/"+key+".html#"+rowName;
					if(imageName!=null) {
						link=writer.createImage(imageName);
						link=writer.createLink(path, link);
					}
					else {
						link=writer.createLink(path,result);
					}
				}
				out.print("<td " +fails+ ">"+link+"</td>");
			}
			writer.endRow();
		}
		
		writer.writeEndTable();
		
		writer.writeDocEnd();
		
		
		
		out.close();
	}

	

	private void writeOutput(File output, List outputData) {
		// TODO Auto-generated method stub
		
	}
	public boolean isVerbose() {
		return verbose;
	}
	public void setVerbose(boolean verbose) {
		this.verbose = verbose;
	}
}
