/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Stack;

import javax.tools.diagnostics.image.Image;
import javax.tools.diagnostics.image.ImageAddressSpace;
import javax.tools.diagnostics.image.ImageProcess;
import javax.tools.diagnostics.runtime.java.JavaRuntime;

import org.apache.commons.jxpath.JXPathContext;
import org.apache.kato.katoview.Output;

/**
 * Provides access to Kato API using JXPath
 * 
 *
 */
public class XPathCommand extends Command {
	private static final String COMMAND_NAME="xpath";
	private static final String DESCRIPTION="JXPath API support";
	private static final String LONG_DESCRIPTION=
		"\"xpath\" command\n"+
		"parameters, space separated:\n"+
		"\t[runtime] <xpath query>\n\n"+
		"The xpath commands executes a JXPath query against the Kato api.\n"+
		"By default, this is from the Image object, with the runtime option\n"+
		"the query is executed against the JavaRuntime object.\n\n"+
		"This example prints out the names of all java threads:\n"+
		"\txpath addressSpaces/currentProcess/runtimes/threads/name\n\n"+
		"Results are prefix with an index e.g. 1:\thello\n\n"
		;
	
	protected XPathCommand(Output o) {
		super(o, COMMAND_NAME, DESCRIPTION, LONG_DESCRIPTION);
		child_commands = null;

	}
	
	public void doCommand(Stack args, Image loadedImage, HashMap properties) {
		String path = "";				
		
		JavaRuntime runtime=null;
		
		if (args.size() == 0 || args.size() > 2) {
			out.println(LONG_DESCRIPTION);
			return;
		} else if (args.size() == 1) {
			path = (String) args.get(0);
			
		} else if (args.size() == 2) {
			if (args.get(1).equals("runtime")) {
				JavaRuntime rt=getJavaRuntime(loadedImage);
				if(rt==null) {
					out.println("Unable to locate Java Runtime");
					return;
					
				}
				runtime=rt;
				path = (String) args.get(0);
			} else {
				out.println(LONG_DESCRIPTION);
				return;
			}
		}
		
		
		JXPathContext context =null;
		if(runtime!=null) {
			context = JXPathContext.newContext(runtime);
		}
		else {
			context=JXPathContext.newContext(loadedImage);
		}
		
		Iterator values = context.iterate(path);

		int count=1;
		while (values.hasNext()) {
			Object value = values.next();
			if(value==null) {
				out.println(count+":\tnull");
			}
			else if (value instanceof String) {
				out.println(count+":\t"+"\""+((String) value)+"\"");		
			} else if (value instanceof Number) {
				out.println(count+":\t"+value.toString());
			} else if (value instanceof Character) {
				out.println(count+":\t"+"'"+((Character)value).charValue()+"'");
			} else {
				out.println(count+":\t"+value.toString());
			}
			count++;
		}
	}

	/**
	 * retrieves first javaruntime from image
	 * 
	 * @param loadedImage
	 * 
	 * @return
	 */
	private JavaRuntime getJavaRuntime(Image loadedImage) {
		
		for(Object space:loadedImage.getAddressSpaces()) {
			if(space instanceof  ImageAddressSpace ) {
				ImageAddressSpace addr=(ImageAddressSpace) space;
				ImageProcess process=addr.getCurrentProcess();
				if(process!=null) {
					for(Object rt:process.getRuntimes()) {
						if(rt instanceof JavaRuntime) return (JavaRuntime) rt;
					}
				}
				
			}
		}
		return null;
	}

}
