/*******************************************************************************
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *   http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ******************************************************************************/
package org.apache.kato.katoview.commands.helpers;

import javax.tools.diagnostics.image.CorruptDataException;
import javax.tools.diagnostics.runtime.java.JavaObject;
import javax.tools.diagnostics.runtime.java.JavaThread;

public class NodeList {

	private MonitorNode head;
	private MonitorNode tail;
	private int id;
	
	public NodeList(MonitorNode node, int _id)
	{
		head = node;
		tail = node;
		node.inList = this;
		id = _id;
	}
	
	public int getID()
	{
		return id;
	}
	
	public boolean equals(Object other)
	{
		return ((NodeList)other).getID() == id;
	}

	// returns null if attached; a NodeList representing the part split off from
	//  this NodeList if a split was required
	public NodeList attachOrSplit(NodeList other, int nodeListNum)
	{
		MonitorNode headToAttach = other.getHead();

		if (tail == headToAttach)
		{
			tail = other.getTail();
			MonitorNode currNode = tail;
			currNode.inList = this;
			if (currNode != headToAttach) {
				do {
					currNode = currNode.waitingOn;
					currNode.inList = this;
				} while (currNode != headToAttach);
			}
			return null;
		}
		else
		{
			MonitorNode currNode = tail;
			NodeList split = new NodeList(tail, nodeListNum);
			do
			{
				if (currNode == headToAttach)
				{
					tail = currNode;
					return split;
				}
				currNode = currNode.waitingOn;
				split.add(currNode);
				if (currNode != head)
					currNode.inList = split;
			} while (currNode != head);
			return null;
		}
	}
	
	public void add(MonitorNode node)
	{
		head = node;
	}
	
	public MonitorNode getHead()
	{
		return head;
	}
	
	public MonitorNode getTail()
	{
		return tail;
	}
	
	public boolean isLoop()
	{
		return tail == head;
	}
	
	// Print this list of monitor nodes (ie the owning threads and there object addresses)
	public String toString()
	{
		String retval = "";
		MonitorNode currNode = tail;
		MonitorNode lastNode = null;
		boolean firstTime = true;
		boolean done = false;
		
		do {
			String name = "";
			String objAddr = "";
			JavaObject object = null;
			JavaThread owner = null;
			
			try {
				owner = currNode.getMonitor().getOwner();
			} catch (CorruptDataException e) {
				name = Exceptions.getCorruptDataExceptionString() + " in owner";
			}

			if (name.equals("")) {
				try {
					name = owner.getName();
				} catch (CorruptDataException e) {
					name = Exceptions.getCorruptDataExceptionString() + " in owner's ID";
				}
			}
			
			object = currNode.getMonitor().getObject();
			if (null == object) {
				objAddr = "raw";
			} else {
				objAddr = Utils.toHex(object.getID().getAddress());
			}
			
			retval += "thread: " + name + " (monitor object: " + objAddr + ") waiting for =>\n\t  ";
			
			lastNode = currNode;
			currNode = currNode.waitingOn;

			if (head == lastNode)
			{
				if (firstTime && head == tail)
					done = false;
				else
					done = true;
			}
			firstTime = false;
			
		} while (!done);
		
		retval = retval.substring(0, retval.length() - 18); // removes the tail of the last entry
		return retval;
	}
}
